#!/usr/bin/env python

import sys
sys.path.append('..')

import numpy as np
from extra_data import RunDirectory, by_id
from pes_to_spec.model import Model, matching_ids

from itertools import product

import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from typing import Optional

from time import time_ns
import pandas as pd

def plot_pes(filename: str, pes_raw_int: np.ndarray):
    """
    Plot low-resolution spectrum.

    Args:
      filename: Output file name.
      pes_raw_int: Low-resolution spectrum.

    """
    fig = plt.figure(figsize=(16, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    ax.plot(pes_raw_int, c='b', lw=3, label="Low-resolution measurement")
    ax.legend()
    ax.set(title=f"",
           xlabel="ToF index",
           ylabel="Intensity")
    fig.savefig(filename)
    plt.close(fig)

def plot_result(filename: str, spec_pred: np.ndarray, spec_smooth: np.ndarray, spec_raw_pe: np.ndarray, spec_raw_int: Optional[np.ndarray]=None):
    """
    Plot result with uncertainty band.

    Args:
      filename: Output file name.
      spec_pred: Predicted result with uncertainty bands in a shape of (3, features).
      spec_smooth: Smoothened expected result with shape (features,).
      spec_raw_pe: x axis with the photon energy in eV.
      spec_raw_int: Original true expected result with shape (features,).

    """
    fig = plt.figure(figsize=(16, 8))
    gs = GridSpec(1, 1)
    ax = fig.add_subplot(gs[0, 0])
    eps = np.mean(spec_pred[:, 1])
    ax.plot(spec_raw_pe, spec_smooth, c='b', lw=3, label="High-resolution measurement (smoothened)")
    ax.plot(spec_raw_pe, spec_pred[:, 0], c='r', lw=3, label="High-resolution prediction")
    ax.fill_between(spec_raw_pe, spec_pred[:, 0] - spec_pred[:, 1], spec_pred[:, 0] + spec_pred[:, 1], facecolor='red', alpha=0.6, label="68% unc. (stat.)")
    ax.fill_between(spec_raw_pe, spec_pred[:, 0] - spec_pred[:, 2], spec_pred[:, 0] + spec_pred[:, 2], facecolor='magenta', alpha=0.6, label="68% unc. (syst., PCA)")
    if spec_raw_int is not None:
        ax.plot(spec_raw_pe, spec_raw_int, c='b', lw=1, ls='--', label="High-resolution measurement")
    ax.legend()
    ax.set(title=f"avg(unc) = {eps}",
           xlabel="Photon energy [eV]",
           ylabel="Intensity")
    fig.savefig(filename)
    plt.close(fig)

def main():
    """
    Main entry point. Reads some data, trains and predicts.
    """
    run_dir = "/gpfs/exfel/exp/SA3/202121/p002935/raw"
    run_id = "r0015"
    # get run
    run = RunDirectory(f"{run_dir}/{run_id}")

    # get train IDs and match them, so we are sure to have information from all needed sources
    # in this example, there is an offset of -2 in the SPEC train ID, so correct for it
    spec_offset = -2
    spec_tid = spec_offset + run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                                 "data.trainId"].ndarray()
    pes_tid = run['SA3_XTD10_PES/ADC/1:network',
                  "digitizers.trainId"].ndarray()
    xgm_tid = run['SA3_XTD10_XGM/XGM/DOOCS:output',
                  "data.trainId"].ndarray()
    # these are the train ID intersection
    # this could have been done by a select call in the RunDirectory, but it would not correct for the spec_offset
    tids = matching_ids(spec_tid, pes_tid, xgm_tid)
    train_tids = tids[:-10]
    test_tids = tids[-10:]

    # read the spec photon energy and intensity
    spec_raw_pe = run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                      "data.photonEnergy"].select_trains(by_id[tids - spec_offset]).ndarray()
    spec_raw_int = run['SA3_XTD10_SPECT/MDL/FEL_BEAM_SPECTROMETER_SQS1:output',
                       "data.intensityDistribution"].select_trains(by_id[tids - spec_offset]).ndarray()

    # read the PES data for each channel
    channels = [f"channel_{i}_{l}"
                for i, l in product(range(1, 5), ["A", "B", "C", "D"])]
    pes_raw = {ch: run['SA3_XTD10_PES/ADC/1:network',
                   f"digitizers.{ch}.raw.samples"].select_trains(by_id[tids]).ndarray()
               for ch in channels}

    # read the XGM information
    #xgm_pressure = run['SA3_XTD10_XGM/XGM/DOOCS', "pressure.pressureFiltered.value"].select_trains(by_id[tids]).ndarray()
    #xgm_pe =  run['SA3_XTD10_XGM/XGM/DOOCS:output', "data.intensitySa3TD"].select_trains(by_id[tids]).ndarray()
    #retvol_raw = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.value"].select_trains(by_id[tids]).ndarray()
    #retvol_raw_timestamp = run["SA3_XTD10_PES/MDL/DAQ_MPOD", "u212.timestamp"].select_trains(by_id[tids]).ndarray()

    t = list()
    t_names = list()

    # these have been manually selected:
    #useful_channels = ["channel_1_D",
    #                  "channel_2_B",
    #                  "channel_3_A",
    #                  "channel_3_B",
    #                  "channel_4_C",
    #                  "channel_4_D"]
    model = Model()

    train_idx = np.isin(tids, train_tids)

    model.debug_peak_finding(pes_raw, "test_peak_finding.png")
    print("Fitting")
    start = time_ns()
    model.fit({k: v[train_idx, :]
               for k, v in pes_raw.items()},
              spec_raw_int[train_idx, :],
              spec_raw_pe[train_idx, :])
    t += [time_ns() - start]
    t_names += ["Fit"]
    spec_smooth = model.preprocess_high_res(spec_raw_int, spec_raw_pe)

    print("Saving the model")
    start = time_ns()
    model.save("model.h5")
    t += [time_ns() - start]
    t_names += ["Save"]

    print("Loading the model")
    start = time_ns()
    model = Model()
    model.load("model.h5")
    t += [time_ns() - start]
    t_names += ["Load"]

    # test
    print("Predict")
    start = time_ns()
    spec_pred = model.predict(pes_raw)
    t += [time_ns() - start]
    t_names += ["Predict"]

    print("Time taken in ms")
    df_time = pd.DataFrame(data=dict(time=t, name=t_names))
    df_time.time *= 1e-6
    print(df_time)

    print("Plotting")
    # plot
    for tid in test_tids:
        idx = np.where(tid==tids)[0][0]
        plot_result(f"test_{tid}.png",
                    spec_pred[idx, :, :],
                    spec_smooth[idx, :],
                    spec_raw_pe[idx, :],
                    spec_raw_int[idx, :])
        for ch in channels:
            plot_pes(f"test_pes_{tid}_{ch}.png", pes_raw[ch][idx, :])

if __name__ == '__main__':
    main()
